-------------------------------------------------------------------------------
--
-- Inverse SBox Module
-- Written for 128-bit AES decryption project
-- Course: CSEE 4840 - Embedded System Design, Spring 2008
-- Authors: Shrivathsa Bhargav (sb2784)
--			Larry Chen (lc2454)
--			Abhinandan Majumdar (am2993)
--			Shiva Ramudit (syr9)
--
-- Last modified: 5-8-2008
-------------------------------------------------------------------------------

library IEEE;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

--Module for inverse S-Box 
entity inv_sbox is
	port (
		it : in std_logic_vector (127 downto 0);
		ot : out std_logic_vector (127 downto 0)
		);
end;

architecture inv_sbox_arch of inv_sbox is
type sbox_row_type is array (0 to 15) of std_logic_vector (7 downto 0);
type sbox_type is array (0 to 15) of sbox_row_type;

signal it00, it01, it02, it03 : std_logic_vector (7 downto 0);
signal it10, it11, it12, it13 : std_logic_vector (7 downto 0);
signal it20, it21, it22, it23 : std_logic_vector (7 downto 0);
signal it30, it31, it32, it33 : std_logic_vector (7 downto 0);
signal ot00, ot01, ot02, ot03 : std_logic_vector (7 downto 0);
signal ot10, ot11, ot12, ot13 : std_logic_vector (7 downto 0);
signal ot20, ot21, ot22, ot23 : std_logic_vector (7 downto 0);
signal ot30, ot31, ot32, ot33 : std_logic_vector (7 downto 0);


--Inverse S-table
constant i_s_box : sbox_type := 
		(
			(x"52", x"09", x"6a", x"d5", x"30", x"36", x"a5", x"38", x"bf", x"40", x"a3", x"9e", x"81", x"f3", x"d7", x"fb"), 
			(x"7c", x"e3", x"39", x"82", x"9b", x"2f", x"ff", x"87", x"34", x"8e", x"43", x"44", x"c4", x"de", x"e9", x"cb"),
			(x"54", x"7b", x"94", x"32", x"a6", x"c2", x"23", x"3d", x"ee", x"4c", x"95", x"0b", x"42", x"fa", x"c3", x"4e"),
			(x"08", x"2e", x"a1", x"66", x"28", x"d9", x"24", x"b2", x"76", x"5b", x"a2", x"49", x"6d", x"8b", x"d1", x"25"),
			(x"72", x"f8", x"f6", x"64", x"86", x"68", x"98", x"16", x"d4", x"a4", x"5c", x"cc", x"5d", x"65", x"b6", x"92"),
			(x"6c", x"70", x"48", x"50", x"fd", x"ed", x"b9", x"da", x"5e", x"15", x"46", x"57", x"a7", x"8d", x"9d", x"84"),
			(x"90", x"d8", x"ab", x"00", x"8c", x"bc", x"d3", x"0a", x"f7", x"e4", x"58", x"05", x"b8", x"b3", x"45", x"06"),
			(x"d0", x"2c", x"1e", x"8f", x"ca", x"3f", x"0f", x"02", x"c1", x"af", x"bd", x"03", x"01", x"13", x"8a", x"6b"),
			(x"3a", x"91", x"11", x"41", x"4f", x"67", x"dc", x"ea", x"97", x"f2", x"cf", x"ce", x"f0", x"b4", x"e6", x"73"),
			(x"96", x"ac", x"74", x"22", x"e7", x"ad", x"35", x"85", x"e2", x"f9", x"37", x"e8", x"1c", x"75", x"df", x"6e"),
			(x"47", x"f1", x"1a", x"71", x"1d", x"29", x"c5", x"89", x"6f", x"b7", x"62", x"0e", x"aa", x"18", x"be", x"1b"),
			(x"fc", x"56", x"3e", x"4b", x"c6", x"d2", x"79", x"20", x"9a", x"db", x"c0", x"fe", x"78", x"cd", x"5a", x"f4"),
			(x"1f", x"dd", x"a8", x"33", x"88", x"07", x"c7", x"31", x"b1", x"12", x"10", x"59", x"27", x"80", x"ec", x"5f"),
			(x"60", x"51", x"7f", x"a9", x"19", x"b5", x"4a", x"0d", x"2d", x"e5", x"7a", x"9f", x"93", x"c9", x"9c", x"ef"),
			(x"a0", x"e0", x"3b", x"4d", x"ae", x"2a", x"f5", x"b0", x"c8", x"eb", x"bb", x"3c", x"83", x"53", x"99", x"61"),
			(x"17", x"2b", x"04", x"7e", x"ba", x"77", x"d6", x"26", x"e1", x"69", x"14", x"63", x"55", x"21", x"0c", x"7d") 
		);
begin

--Splitting 128 bit bus into individual 8 bit lines
it00 <= it(127 downto 120);
it01 <= it(119 downto 112);
it02 <= it(111 downto 104);
it03 <= it(103 downto 96);
it11 <= it(95 downto 88);
it12 <= it(87 downto 80);
it13 <= it(79 downto 72);
it10 <= it(71 downto 64);
it22 <= it(63 downto 56);
it23 <= it(55 downto 48);
it20 <= it(47 downto 40);
it21 <= it(39 downto 32);
it33 <= it(31 downto 24);
it30 <= it(23 downto 16);
it31 <= it(15 downto 8);
it32 <= it(7 downto 0);



--Replacing the data from Inverse S-box
ot00 <= i_s_box(conv_integer(it00(7 downto 4)))(conv_integer(it00(3 downto 0)));
ot01 <= i_s_box(conv_integer(it01(7 downto 4)))(conv_integer(it01(3 downto 0)));
ot02 <= i_s_box(conv_integer(it02(7 downto 4)))(conv_integer(it02(3 downto 0)));
ot03 <= i_s_box(conv_integer(it03(7 downto 4)))(conv_integer(it03(3 downto 0)));
ot10 <= i_s_box(conv_integer(it10(7 downto 4)))(conv_integer(it10(3 downto 0)));
ot11 <= i_s_box(conv_integer(it11(7 downto 4)))(conv_integer(it11(3 downto 0)));
ot12 <= i_s_box(conv_integer(it12(7 downto 4)))(conv_integer(it12(3 downto 0)));
ot13 <= i_s_box(conv_integer(it13(7 downto 4)))(conv_integer(it13(3 downto 0)));
ot20 <= i_s_box(conv_integer(it20(7 downto 4)))(conv_integer(it20(3 downto 0)));
ot21 <= i_s_box(conv_integer(it21(7 downto 4)))(conv_integer(it21(3 downto 0)));
ot22 <= i_s_box(conv_integer(it22(7 downto 4)))(conv_integer(it22(3 downto 0)));
ot23 <= i_s_box(conv_integer(it23(7 downto 4)))(conv_integer(it23(3 downto 0)));
ot30 <= i_s_box(conv_integer(it30(7 downto 4)))(conv_integer(it30(3 downto 0)));
ot31 <= i_s_box(conv_integer(it31(7 downto 4)))(conv_integer(it31(3 downto 0)));
ot32 <= i_s_box(conv_integer(it32(7 downto 4)))(conv_integer(it32(3 downto 0)));
ot33 <= i_s_box(conv_integer(it33(7 downto 4)))(conv_integer(it33(3 downto 0)));

-- Putting individual 8 bit lines to a 128 bit wide bus
ot <= ot00 & ot01 & ot02 & ot03 &
		 ot10 & ot11 & ot12 & ot13 &
		 ot20 & ot21 & ot22 & ot23 &
		 ot30 & ot31 & ot32 & ot33 ;

end inv_sbox_arch;