-------------------------------------------------------------------------------
--
-- SD/MMC interface (SPI-style) for the Apple ][ Emulator
-- Written for 128-bit AES decryption project
-- Course: CSEE 4840 - Embedded System Design, Spring 2008
-- Authors: Shrivathsa Bhargav (sb2784)
--			Larry Chen (lc2454)
--			Abhinandan Majumdar (am2993)
--			Shiva Ramudit (syr9)
--
-- Originally authored by Stephen A. Edwards (sedwards@cs.columbia.edu)
--
-- Last modified: 5-8-2008
-------------------------------------------------------------------------------

library IEEE;
use IEEE.STD_LOGIC_1164.all;
use ieee.numeric_std.all;

entity spi_controller is
  
  port (

	-- Bus interface signals
	read  		: in std_logic;
	write 		: in std_logic;
	chipselect 	: in std_logic;
	address		: in unsigned (23 downto 0);
	readdata	: out unsigned (31 downto 0);
	writedata	: in unsigned (31 downto 0);
	-- End of bus interface

    CS_N           : out std_logic := '1';  	-- MMC chip select
    MOSI           : out std_logic;     		-- Data to card (master out slave in)
    MISO           : in  std_logic;   			-- Data from card (master in slave out)
    SCLK           : out std_logic;     		-- Card clock
    
    clk      	   : in  std_logic;    			-- System clock
    reset          : in  std_logic);

end spi_controller;

architecture rtl of spi_controller is

  type states is (RESET_STATE,
                  RESET_CLOCKS1,
                  RESET_CLOCKS2,
                  RESET_SEND_CMD0,
                  RESET_SEND_CMD1,
                  RESET_CHECK_CMD1,
                  RESET_SEND_SET_BLOCKLEN,
                  IDLE,
                  READ_BLOCK,
                  READ_BLOCK_WAIT,
                  READ_BLOCK_DATA,
                  READ_BLOCK_CRC,
                  SEND_CMD,
                  RECEIVE_BYTE_WAIT,
                  RECEIVE_BYTE
);

  signal state : states := RESET_STATE;
  signal return_state : states;
  signal sclk_sig : std_logic := '0';
  signal counter : unsigned(7 downto 0) := (others => '0');
  signal byte_counter : unsigned(31 downto 0) := (others => '0');
  signal command : unsigned(55 downto 0) := (others => '1');
  signal recv_byte : unsigned(7 downto 0);
  signal CRC_byte	: std_logic := '0';
  signal start    	: std_logic := '0';
  signal addr	    : unsigned(31 downto 0) := (others => '0');
  signal eor		: std_logic;
  signal rec32bit	: unsigned(31 downto 0);
  signal bufAddr	: integer;
  signal bufResult	: unsigned (31 downto 0);
  signal bufCounter	: unsigned(8 downto 0) := "000000000";
  
  type ram_type is array (0 to 511) of unsigned(7 downto 0);
  signal RAM : ram_type := (others => (others => '0'));
    
begin

  SCLK <= sclk_sig;

	process (clk)
	begin
		if rising_edge(clk) then
			if reset = '1' then
				readdata <= (others => '0');
			else
				if chipselect = '1' then
					if address(0) = '1' then
						if write = '1' then					
							start <=  writedata(0);
						end if;
					elsif address(1) = '1' then
						if write = '1' then
							addr <= writedata;
						end if;
					elsif address(4) = '1' then
						if read = '1' then
							readdata <= "0000000000000000000000000000000" & eor;
						end if;
					elsif address(6) = '1' then
						if write = '1' then
							bufAddr <= to_integer(writedata);
						end if;	
					elsif address(7) = '1' then	
						if read = '1' then
							readdata <= bufResult;
						end if;
					end if;
				end if;
			end if;
		end if;
	end process;

-- Give the data to the processor in 32 bit blocks
bufResult <= RAM(bufAddr) & RAM(bufAddr + 1) & RAM(bufAddr + 2) & RAM(bufAddr + 3);

  fsm_ff : process(clk)
  begin
    if rising_edge(clk) then
      if reset = '1' then
        state <= RESET_STATE;
        sclk_sig <= '0';
        CS_N <= '1';
        command <= (others => '1');
        counter <= (others => '0');
        byte_counter <= (others => '0');
		eor <= '0';
      else		
        case state is

          when RESET_STATE =>
            counter <= TO_UNSIGNED(160,8);
            state <= RESET_CLOCKS1;

            -- Output a series of clock signals to wake up the chip
          when RESET_CLOCKS1 =>
            if counter = 0 then
              counter <= TO_UNSIGNED(32, 8);
              CS_N <= '0';
              state <= RESET_CLOCKS2;
            else
              counter <= counter - 1;
              sclk_sig <= not sclk_sig;
            end if;
            
          when RESET_CLOCKS2 =>
            if counter = 0 then
              state <= RESET_SEND_CMD0;
            else
              counter <= counter - 1;
              sclk_sig <= not sclk_sig;
            end if;

            -- Send CMD0 : GO_IDLE_STATE
          when RESET_SEND_CMD0 =>
            command <= x"FF400000000095";
            counter <= TO_UNSIGNED(55, 8);
            return_state <= RESET_SEND_CMD1;
            state <= SEND_CMD;

            -- Send CMD1 : SEND_OP_CMD
          when RESET_SEND_CMD1 =>
            command <= x"FF410000000001";
            counter <= TO_UNSIGNED(55, 8);
            return_state <= RESET_CHECK_CMD1;
            state <= SEND_CMD;

		    -- Wait for a response from the card
          when RESET_CHECK_CMD1 =>
            if recv_byte = x"00" then
              state <= RESET_SEND_SET_BLOCKLEN;
            else
             state <= RESET_SEND_CMD1;              
            end if;

			-- Set the block length of the card to 512 bytes
          when RESET_SEND_SET_BLOCKLEN =>		
     		command <= x"FF500000020001";  -- CMD16: SET_BLOCKLEN (512 bytes)
            counter <= TO_UNSIGNED(55, 8);
            return_state <= IDLE;
            state <= SEND_CMD;

			-- Idle until a start signal is sent
          when IDLE =>
			eor <= '1';
            if start = '1' then
              state <= READ_BLOCK;
              eor <= '0';
            end if;

          -- Set the block read starting address
          when READ_BLOCK =>
            command <= x"FF51" & addr & x"01";  -- READ_SINGLE_BLOCK
            counter <= TO_UNSIGNED(55, 8);
            return_state <= READ_BLOCK_WAIT;
            state <= SEND_CMD;

            -- Wait for a 0 to signal the start of the block,
            -- then read the first byte
          when READ_BLOCK_WAIT =>
            if sclk_sig = '1' and MISO = '0' then
				state <= READ_BLOCK_DATA;
				byte_counter <= x"000001FF";	-- Set the byte counter to 511 [0-> 511]
				bufCounter <= (others => '0');
				counter <= TO_UNSIGNED(7, 8);
            end if;
            sclk_sig <= not sclk_sig;            

			-- Read bytes from the data until all the bytes have been read
			-- If all bytes are read then we enter the READ_BLOCK_CRC state
          when READ_BLOCK_DATA =>
            if byte_counter = x"00" then
              counter <= TO_UNSIGNED(7, 8);
              return_state <= READ_BLOCK_CRC;
              state <= RECEIVE_BYTE;
            else
              byte_counter <= byte_counter - 1;
              return_state <= READ_BLOCK_DATA;
              counter <= TO_UNSIGNED(7, 8);
              state <= RECEIVE_BYTE;
            end if;

          when READ_BLOCK_CRC =>
            counter <= TO_UNSIGNED(15, 8);
			CRC_byte <= '1';	
            return_state <= IDLE;
            state <= RECEIVE_BYTE;

          -- Send the command.  Set counter=54 and return_state first
          when SEND_CMD =>
            if sclk_sig = '1' then
              if counter = 0 then
                state <= RECEIVE_BYTE_WAIT;
              else
                counter <= counter - 1;
				command <= command(54 downto 0) & "1";
              end if;
            end if;
            sclk_sig <= not sclk_sig;

          -- Wait for a "0", indicating the first bit of a response.
          -- Set return_state first
          when RECEIVE_BYTE_WAIT =>
            if sclk_sig = '1' then
              if MISO = '0' then
                recv_byte <= (others => '0');
                counter <= TO_UNSIGNED(6, 8);  -- Already read bit 7
                state <= RECEIVE_BYTE;
              end if;
            end if;
            sclk_sig <= not sclk_sig;

          -- Receive a byte.  Set counter to 7 and return_state before entry
          when RECEIVE_BYTE =>
            if sclk_sig = '1' then
              recv_byte <= recv_byte(6 downto 0) & MISO;
              if counter = 0 then
                state <= return_state;
				-- Buffer all the data in RAM
				if CRC_byte = '0' then
					RAM(to_integer(bufCounter)) <= recv_byte(6 downto 0) & MISO;
					bufCounter <= bufCounter + 1;
				-- Ignore bytes read after data block
				else
					eor <= '1';
					CRC_byte <= '0';
				end if;
              else
                counter <= counter - 1;
              end if;
            end if;
            sclk_sig <= not sclk_sig;
                        
          when others => null;
        end case;
      end if;
    end if;
  end process fsm_ff;

-- Data sent to the SD Card, MSB first
MOSI <= command(55);  

end rtl;

